@@ -84,8 +84,10 @@ void DeclarationBuilder::setPrebuilding(bool prebuilding)
     m_prebuilding = prebuilding;
 }
 
-ReferencedTopDUContext DeclarationBuilder::build(const IndexedString& url, Ast* node, ReferencedTopDUContext updateContext)
+ReferencedTopDUContext DeclarationBuilder::build(const IndexedString& url, Ast* node,
+                                                 const ReferencedTopDUContext& updateContext_)
 {
+    ReferencedTopDUContext updateContext(updateContext_);
     m_correctionHelper.reset(new CorrectionHelper(url, this));
 
     // The declaration builder needs to run twice, so it can resolve uses of e.g. functions
@@ -362,7 +364,21 @@ void DeclarationBuilder::visitWithItem(WithItemAst* node)
         // For statements like "with open(f) as x", a new variable must be created; do this here.
         ExpressionVisitor v(currentContext());
         v.visitNode(node->contextExpression);
-        visitVariableDeclaration<Declaration>(node->optionalVars, nullptr, v.lastType());
+        auto mgrType = v.lastType();
+        auto enterType = mgrType; // If we can't find __enter__(), assume it returns `self` like file objects.
+
+        static const IndexedIdentifier enterId(KDevelop::Identifier("__enter__"));
+
+        DUChainReadLocker lock;
+        if ( auto enterFunc = dynamic_cast<FunctionDeclaration*>(
+                Helper::accessAttribute(mgrType, enterId, topContext()))) {
+            if ( auto enterFuncType = enterFunc->type<FunctionType>() ) {
+                enterType = enterFuncType->returnType();
+            }
+        }
+        lock.unlock();
+        // This may be any assignable expression, e.g. `with foo() as bar[3]: ...`
+        assignToUnknown(node->optionalVars, enterType);
     }
     Python::AstDefaultVisitor::visitWithItem(node);
 }
@@ -908,7 +924,7 @@ void DeclarationBuilder::addArgumentTypeHints(CallAst* node, DeclarationPointer
     // - 'parameters' refers to the parameters of the function definition.
     // - 'arguments' refers to the arguments of the function call.
 
-    DUContext* parameterContext = DUChainUtils::getArgumentContext(function);
+    DUContext* parameterContext = DUChainUtils::argumentContext(function);
     FunctionType::Ptr functionType = function->type<FunctionType>();
     if ( ! parameterContext || ! functionType ) {
         return;
@@ -1341,6 +1357,7 @@ void DeclarationBuilder::visitAnnotationAssignment(AnnotationAssignmentAst* node
 void DeclarationBuilder::visitClassDefinition( ClassDefinitionAst* node )
 {
     visitNodeList(node->decorators);
+    visitNodeList(node->baseClasses);
     const CorrectionHelper::Recursion r(m_correctionHelper->enterClass(node->name->value));
 
     StructureType::Ptr type(new StructureType());
@@ -1508,7 +1525,7 @@ void DeclarationBuilder::visitFunctionDefinition( FunctionDefinitionAst* node )
     }
 
     if ( ! dec->isStatic() ) {
-        DUContext* args = DUChainUtils::getArgumentContext(dec);
+        DUContext* args = DUChainUtils::argumentContext(dec);
         if ( args )  {
             QVector<Declaration*> parameters = args->localDeclarations();
             static IndexedString newMethodName("__new__");
